//Copyright 1998 President and Fellows of Harvard University
//motif.cpp

#include "motif.h"
#include "hack.h"

void Motif::usage_error(){
    cerr<<"ScanACE version 1.3  October 18, 1999\n";
    cerr<<"Usage:  ScanACE -i (Alignace output file for input)\n";
    cerr<<"-n (number of motif to scan)\n";
    cerr<<"   if this parameter is omitted, all motifs in the file are scanned.\n";
    cerr<<"-c (number of standard deviations below the average score for cutoff, def. 0\n";
    cerr<<"-m (max number of sites to return, default 10000)\n";
    cerr<<"-s (return the best # of sites, overrides other options)\n";
    cerr<<"-a <string> orftable filename\n";
    cerr<<"-b <string> non-orf features table filename\n";
    cerr<<"-z <string> genome filename\n";
    cerr<<"-x ignore genome and features information\n";
    cerr<<"-g genome gc content (default 0.5)\n";
    cerr<<"-y use default yeast values for -a, -b, -z options,i.e.,\n";
    cerr<<"  -z yeast.seq -a ORF_table.txt -b other_features_table.txt\n";
    cerr<<"  may be partially overridden, e.g., -y -a -b is the same as -z yeast.seq\n";
    cerr<<"-o (output filename base) defaults to input file name\n";
    cerr<<"  output files are named basename_n#.scn, where # is the motif number\n";
    cerr<<"To create a non-AlignACE file for input, make a text file with the line\n";
    cerr<<"'Motif 1' at the top followed by a blank line followed by lines of\n";
    cerr<<"aligned sequence constituting the motif.  The last line of the file should be\n";
    cerr<<"a string of *'s, one under each active column in the motif.  For example,\n";
    cerr<<"Motif 1\n\n";
    cerr<<"ccgaatggcc\ncctcagggcc\nccggtgggcc\nctgtatggcc\ncagaatggcc\n*** ******\n";
    exit(0);
}


void Motif::freq_matrix_fast(double *fm, const bool useactive, const int five, const int three){
  //useactive defaults to true, means that only the columns will be considered
  int matsize=wide();
  if(useactive) matsize=ncols();
  matsize+=five+three;
  matsize*=4;
  int n;
  for(n=0;n<matsize;n++){
    fm[n]=0;
  }
  list<Site>::iterator p;
  list<bool>::iterator a;
  list<bool>::reverse_iterator b;
  for(p=sites.begin();p!=sites.end();p++){
    //for each site
    int x=(*p).seqnum();
    int y=(*p).pos();//seq_fast[x][y] is the first position
    int i;
    if((*p).strand()){
      y-=five;
      n=0;
      for(i=0;i<five;n+=4,y++,i++){
	if(seq_fast[x][y]==-1){
	  fm[n]+=(1-motif_gc)/2;fm[n+3]+=(1-motif_gc)/2;
	  fm[n+1]+=motif_gc/2;fm[n+2]+=motif_gc/2;
	}
	else fm[n+seq_fast[x][y]]++;
      }
      a=active.begin();
      for(i=0;i<wide();y++,a++,i++){
	if(!useactive||*a){
	  if(seq_fast[x][y]==-1){
	    fm[n]+=(1-motif_gc)/2;fm[n+3]+=(1-motif_gc)/2;
	    fm[n+1]+=motif_gc/2;fm[n+2]+=motif_gc/2;
	  }
	  else fm[n+seq_fast[x][y]]++;
	  n+=4;
	}
      }
      for(i=0;i<three;n+=4,y++,i++){
	if(seq_fast[x][y]==-1){
	  fm[n]+=(1-motif_gc)/2;fm[n+3]+=(1-motif_gc)/2;
	  fm[n+1]+=motif_gc/2;fm[n+2]+=motif_gc/2;
	}
	else fm[n+seq_fast[x][y]]++;
      }
    }
    else{//reverse strand
      //switch five and three, run matrix and active in reverse,
      //complement implicit
      y-=three;
      n=matsize-1;
      for(i=0;i<three;n-=4,y++,i++){
	if(seq_fast[x][y]==-1){
	  fm[n]+=(1-motif_gc)/2;fm[n-3]+=(1-motif_gc)/2;
	  fm[n-1]+=motif_gc/2;fm[n-2]+=motif_gc/2;
	}
	else fm[n-seq_fast[x][y]]++;
      }
      b=active.rbegin();
      for(i=0;i<wide();y++,b++,i++){
	if(!useactive||*b){
	  if(seq_fast[x][y]==-1){
	    fm[n]+=(1-motif_gc)/2;fm[n-3]+=(1-motif_gc)/2;
	    fm[n-1]+=motif_gc/2;fm[n-2]+=motif_gc/2;
	  }
	  else fm[n-seq_fast[x][y]]++;
	  n-=4;
	}
      }
      for(i=0;i<five;n-=4,y++,i++){
	if(seq_fast[x][y]==-1){
	  fm[n]+=(1-motif_gc)/2;fm[n-3]+=(1-motif_gc)/2;
	  fm[n-1]+=motif_gc/2;fm[n-2]+=motif_gc/2;
	}
	else fm[n-seq_fast[x][y]]++;
      }
    }
  }
}


void Motif::create_matrix(){

  int matsize=4*ncols();
  double *fm=new double[matsize];
  freq_matrix_fast(fm);
  
  int i=0;
  double gc;
  //double sum=number()+0.5;
  double sum=number()+1.0;
  for(i=0;i<matsize;i++){
    if(i%4==0||i%4==3) gc=(1.0-motif_gc)/2.0;
    else gc=motif_gc/2.0;
    //mat_fast[i]=log(double(fm[i]+0.5)/sum/gc)/log(2.0);
    mat_fast[i]=log((fm[i]+gc)/sum/gc)/log(2.0);
    //this now corresponds exactly to the log of the score used by AlignACE
  }
  delete[] fm;
}

double Motif::score(const string &site){
  //only give it the active columns
  if(updated) update();
  string::const_iterator s=site.begin();
  double score=0.0;
  static map<char,int> convert;
  static bool init=false;
  if(!init){
    convert['a']=convert['A']=0;
    convert['c']=convert['C']=1;
    convert['g']=convert['G']=2;
    convert['t']=convert['T']=3;
    convert['n']=convert['N']=-1;
    init=true;
	}
  for(int i=0;s!=site.end();s++,i+=4){
    if(convert[*s]!=-1) score+=mat_fast[i+convert[*s]];
    else{
      score+=(1-motif_gc)*(mat_fast[i]+mat_fast[i+3]);
      score+=motif_gc*(mat_fast[i+1]+mat_fast[i+2]);
    }
  }
  return score;
}

double Motif::score_raw(const string &site){
  //this version takes raw sequence and only looks at active columns
  if(updated) update();
  string::const_iterator s=site.begin();
  double score=0.0;
  static map<char,int> convert;
  static bool init=false;
  if(!init){
    convert['a']=convert['A']=0;
    convert['c']=convert['C']=1;
    convert['g']=convert['G']=2;
    convert['t']=convert['T']=3;
    convert['n']=convert['N']=-1;
    init=true;
  }
  list<bool>::const_iterator a=active.begin();
  for(int i=0;s!=site.end();s++,a++){
    if(*a){
      if(convert[*s]!=-1) score+=mat_fast[i+convert[*s]];
      else{
	score+=(1-motif_gc)*(mat_fast[i]+mat_fast[i+3]);
	score+=motif_gc*(mat_fast[i+1]+mat_fast[i+2]);
      }
      i+=4;
    }
  }
  return score;
}

double Motif::score_partial(const string &site, int offset, int &num_cols_used){
  //this version takes raw sequence and only looks at active columns
  //the site is presumably shorter than the matrix in this version.  The offset is used to
  //pick which portion of the matrix to use to score the site, i.e., for offset=2,
  //  Site:    ACGTACGTACGT
  //Matrix:  123456789012345678
  //num_cols_used returns the number of active columns encountered in the scoring
  num_cols_used=0;
  if(updated) update();
  string::const_iterator s=site.begin();
  double score=0.0;
  static map<char,int> convert;
  static bool init=false;
  if(!init){
    convert['a']=convert['A']=0;
    convert['c']=convert['C']=1;
    convert['g']=convert['G']=2;
    convert['t']=convert['T']=3;
    convert['n']=convert['N']=-1;
    init=true;
  }
  if(site.size()>wide()-offset){
    cerr<<"Improper input to Motif::score_partial(), site too wide for matrix.\n";
    exit(0);
  }
  list<bool>::const_iterator a=active.begin();
  int i=0;
  for(int j=0;j<offset;j++,a++){
    if(*a) i+=4;
  }
  for(;s!=site.end();s++,a++){
    if(*a){
      if(convert[*s]!=-1) score+=mat_fast[i+convert[*s]];
      else{
	score+=(1-motif_gc)*(mat_fast[i]+mat_fast[i+3]);
	score+=motif_gc*(mat_fast[i+1]+mat_fast[i+2]);
      }
      i+=4;
      num_cols_used++;
    }
  }
  return score;
}


double Motif::score(const short *s){
  //takes a pointer into seq_fast, should consider only active columns
  if(updated) update();
  double score=0.0;
  list<bool>::const_iterator a=active.begin();
  for(int i=0,j=0;j!=wide();j++,a++){
    if(*a){
      if(s[j]!=-1) score+=mat_fast[i+s[j]];
      else{
	score+=(1-motif_gc)*(mat_fast[i]+mat_fast[i+3]);
	score+=motif_gc*(mat_fast[i+1]+mat_fast[i+2]);
      }
      i+=4;
    }
  }
  return score;
}

double Motif::score_rev(const short *s){
  //takes a pointer into seq_fast, should consider only active columns
  if(updated) update();
  list<bool>::reverse_iterator a=active.rbegin();
  double score=0.0;
  for(int i=4*ncols()-1,j=0;j!=wide();j++,a++){
    if(*a){
      if(s[j]!=-1) score+=mat_fast[i-s[j]];
      else{
	score+=(1-motif_gc)*(mat_fast[i]+mat_fast[i-3]);
	score+=motif_gc*(mat_fast[i-1]+mat_fast[i-2]);
      }
      i-=4;
    }
  }
  return score;
}


Motif::Motif(const Motif& m):Sites(m){
	mat_fast=new double[4*m.ncols()];
	for(int i=0;i<4*m.ncols();i++){
		mat_fast[i]=m.mat_fast[i];
	}
}

Motif& Motif::operator= (const Motif& m){
	if(this!=&m){	
		delete[] mat_fast;
		mat_fast=new double[4*m.ncols()];
		for(int i=0;i<4*m.ncols();i++){
			mat_fast[i]=m.mat_fast[i];
		}
	}
	return(*this);
}


bool MatrixTest::operator()(const short *s,bool t){
	double sc;
	if(t) sc=mot.score(s);
	else sc=mot.score_rev(s);
	sum+=sc;
	sum_squares+=sc*sc;
	num_calls++;
	return sc>cutoff;
}


void Motif::put_motif(ostream &fout){
	sites.sort();
	list<Site>::iterator s=sites.begin();
	for(;s!=sites.end();s++){
		fout<<seq(*s,false)<<'\t'<<(*s).seqnum()<<'\t'<<
			(*s).pos()<<'\t'<<(*s).strand()<<'\t'<<
			score(*s)<<'\n';
	}
	list<bool>::iterator a=active.begin();
	for(;a!=active.end();a++){
		if(*a) fout<<'*';
		else fout<<' ';
	}
	fout <<'\n';
}


void Motif::stats_all(double &avg, double &sd){
	MatrixTest mt(*this,DBL_MAX);
	add_sites_better(mt);
	//no sites should be added, but the scores should accumulate
	avg=mt.avg_scores();
	sd=mt.sd_scores();
}

void Motif::stats_self(double &avg, double &sd){
	avg=0.0;sd=0.0;
	list<Site>::const_iterator ls=sites.begin();
	for(;ls!=sites.end();ls++){
		double sc=score(*ls);
		avg+=sc;
		sd+=sc*sc;
	}
	avg=avg/number();
	sd=(sd-avg*avg*number())/(number()-1);
	if(sd>0.0) sd=sqrt(sd);
	else sd=0.0;
}


void put_score_sites(Genome &gn,Motif &m,ostream &fout, int cut, int max, int top_sites, bool locinfo){
  double mot_avg,mot_sd;
  m.stats_self(mot_avg,mot_sd);
  double cutoff = mot_avg - cut * mot_sd;
  fout<<"Motif Average: "<<mot_avg<<'\t'<<"Std. Dev.: "<<mot_sd<<'\n'<<'\n';
  
  MatrixTest mt(m,cutoff);
  gn.clear_sites();
  gn.copy_column_stx(m);
  gn.add_sites_better(mt,false);//one site per strand
  if(top_sites<=0) fout<<"There are "<<gn.number()<<" sites scoring better than this cutoff.\n\n";
  if(top_sites>0){
    if(!(mot_sd>.01)){
      //cerr<<"Std. dev. is too low.  Using another value to modify cutoffs.\n";
      mot_sd=cutoff/3.0;
      if(mot_sd<0) mot_sd=-mot_sd;
    }
    while(gn.number()<top_sites){//need more sites to look at
      cutoff-=mot_sd;
      //cerr<<"New cutoff = "<<cutoff<<"\n";
      mt.set_cutoff(cutoff);
      gn.clear_sites();
      gn.add_sites_better(mt,false);//can add sites on both strands
    }
  }
  
  int num=gn.number();
  if(top_sites<0&&num>max){
    num=max;
    fout<<"The output is being truncated at "<<max<<" sites.  Some equally valid sites will not be listed.\n";
  }
  fout<<"Sites:\n";
  map<int,Site> sm;
  Heap hp(num,3);
  int item=1;
  list<Site> ls=gn.where();
  list<Site>::iterator hit=ls.begin();
  for(;hit!=ls.end();hit++){
    double sw=m.score(gn.seq(*hit));
    //double sc=m.score(reverse_comp(gn.seq(*hit)));
    //if(sc>sw) (*hit).switch_strand();
    sm[item]=*hit;
    double sbest=sw;
    //if(sc>sw)sbest=sc;
    hp.insert(item,-sbest);//use negative to make the best the least
    item++;
    if(item>num) break;
  }
  double lastscore=-9999.0;
  for(int f=0;;f++){
    item=hp.del_min();
    if(item==-1) break;
    Site ps=sm[item];		
    if(top_sites>0&&f>=top_sites&&m.score(gn.seq(ps))!=lastscore) break;
    fout<<gn.keyname((ps).seqnum())<<'\t'
	<<(ps).pos()<<'\t'
	<<(ps).strand()<<'\t'
	<<gn.seq(ps,false)<<'\t'
      //	<<(ps).seqnum()<<'\t'
	<<m.score(gn.seq(ps))
	<<'\n';
    if(locinfo){
      OrfLocale oloc(ps,m.wide(),gn);
      fout<<oloc.desc_long()<<'\n';
    }
    lastscore=m.score(gn.seq(ps));
  }
}

void score_sites(int argc, char* argv[]){
  if(argc<2){
    Motif::usage_error();
  }
  string i,o;
  GetArg2(argc,argv,"-i",i);
  GetArg2(argc,argv,"-o",o);
  
  int n=0,c=0,m=10000,s=-1;
  GetArg2(argc,argv,"-n",n);
  GetArg2(argc,argv,"-c",c);
  GetArg2(argc,argv,"-m",m);
  GetArg2(argc,argv,"-s",s);
  
  string a,b,z;
  if(GetArg2(argc,argv,"-y")){
    a="ORF_table.txt";
    b="other_features_table.txt";
    z="yeast.seq";
  }
  GetArg2(argc,argv,"-a",a);//orf table
  GetArg2(argc,argv,"-b",b);//non-orf table
  GetArg2(argc,argv,"-z",z);//genome

  double g=0.5;
  GetArg2(argc, argv, "-g", g);

  //if no output file base specified, use input name
  if(o.length()==0){
    if(i.size()!=0) o=i.substr(0,i.find_last_of('.'));
    else{
      cerr<<"No input file specified.\n";
      exit(0);
    }
  }
  o+="_n";

  string cmdline;
  for (int arg=0;arg<argc;arg++) {
    cmdline+=(argv[arg]);
    cmdline+=" ";
  }

  ifstream fin(i.c_str());
  if(!fin){
    cerr<<"Could not open file "<<i<<'\n';
    exit(0);
  }

  int n1,n2;
  if(n>0){
    n1=n2=n;
  }
  else{
    n1=1;
    n2=number_motifs(i.c_str());
  }

  vector<string> genseq,nameset;
  get_fasta_fast(z.c_str(),genseq,nameset);
  Genome gen(genseq,nameset);

  if(!(GetArg2(argc,argv,"-x"))){
    if(a.size()!=0){
      ifstream tab_orf(a.c_str());
      gen.load_table(tab_orf,true);
    }
    if(b.size()!=0){
      ifstream tab_nonorf(b.c_str());
      gen.load_table(tab_nonorf,false);
    }
    gen.add_ends(genseq);
  }

  for(int k=n1;k<=n2;k++){
    Motif mot=read_alignace_motif(fin,k);
    if(mot.number()<1){
      cerr<<"Not enough sites to bother searching.\n";
      exit(0);
    }
    mot.set_gc(g);

    string of=o+int_to_str(k);
    of+=".scn";
    //cerr<<of<<'\t'<<cmdline<<'\n';
    ofstream fout(of.c_str());
    fout<<"ScanACE version 1.3  October 18, 1999\n";
    fout<<cmdline<<'\n';
    fout<<"Input Motif:\n\n";
    mot.put_motif(fout);
  
    if(!(GetArg2(argc,argv,"-x"))){
      put_score_sites(gen,mot,fout,c,m,s,true);
    }
    else{
      put_score_sites(gen,mot,fout,c,m,s,false);
    }
  }
}


Motif read_alignace_motif(istream &fin, int num){
  //reads in the motif number num and creates a Motif with the appropriate
  //column structure as detailed by a string of asterisks following the sequences
  string s;
  int x;
  while (getline(fin,s,'\n')){
    if(s.size()<5) continue;
    if(s.substr(0,5)=="Motif"){
      if(s.size()>=12&&s.substr(0,12)=="Motif number"){
	x=str_to_int(s.substr(13));
      }
      else{
	x=str_to_int(s.substr(6));
      }
      if(x==num){
	break;
      }
    }
  }
  vector<string> seqset;
  while(getline(fin,s,'\n')){
    if(s.size()<2) continue;
    if(s.find('*')!=string::npos) break;
    s = s.substr(0,s.find('\t'));
    //this sequence is pre- and post- padded with spaces for equal length sequences
    seqset.push_back(s);
  }
  if(s.find('*')==string::npos){
    cerr<<"problem with input to read_alignace_motif():\n";
    cerr<<" motif "<<num<<" column structure not specified\n";
    exit(0);
  }
  Motif mot(seqset);
  mot.set_columns(s);
  for(int k=1;k<=seqset.size();k++){
    mot.add(Site(k,1,true));
  }
  mot.trim_column_stx();
  return mot;
}
