//Copyright 1998 President and Fellows of Harvard University
//orfs.h

#ifndef _orfs
#define _orfs
#include "standard.h"
#include "sites.h"

class Orf{
  int chrom,beg,end;
  bool watson;
  string name1;//y-name
  string name2;//gene name
  string desc1;//exon structure
  string desc2;//known function
  vector<string> category;
  bool isorf;
 public:
  Orf():chrom(-1),beg(-1),end(-1),watson(false){}
  Orf(int c, int b, int e, bool s):chrom(c),beg(b),end(e),watson(s),isorf(true){}
  void set_name1(string s){name1=s;}
  void set_name2(string s){name2=s;}
  void set_desc1(string s){desc1=s;}
  void set_desc2(string s){desc2=s;}
  void add_funcat(string f){if(!in_funcat(f)) category.push_back(f);}
  void set_nonorf(){isorf=false;}
  int seqnum() const {return chrom;}
  int left() const {return beg;}
  int right() const {return end;}
  int start() const {if(watson) return beg;return end;}
  int stop() const {if(watson) return end;return beg;}
  int length() const {return (end-beg+1);}
  bool strand() const {return watson;}
  string yname() const {return name1;}
  string genename() const {return name2;}
  string description() const {return (desc1+" "+desc2);}
  string full_desc() const {return (name1+" "+name2+" "+desc1+" "+desc2);}
  string gene_desc() const {return desc2;}
  vector<string> funcat() const {return category;}
  bool is_orf() const {return isorf;}
  bool is_null() const {if(chrom==-1) return true;return false;}
  void set_null() {chrom=-1;}
  bool is_end() const {if(desc2=="end") return true;return false;}
  bool in_funcat(string s);
};

bool operator< (const Orf s, const Orf t);
bool operator== (const Orf s, const Orf t);
// Equality for sites depends on chrom, then beg, then end.
bool operator!= (const Orf s, const Orf t);
bool operator> (const Orf s, const Orf t);

bool nested_orfs(const Orf &o, const Orf &p);
bool overlapped_orfs(const Orf &o, const Orf &p);
int overlap_size(const Orf &o, const Orf &p);


class Orfs{
  list<Orf> *orf_list;
  int num_sequences;
  bool mapped;
  bool sorted;
  map<string,Orf> lookup;
 public:
  Orfs(int num){orf_list=new list<Orf>[num+1];num_sequences=num;
  mapped=false;sorted=false;}
  ~Orfs(){delete[] orf_list;}
  Orfs(const Orfs&);
  Orfs& operator= (const Orfs&);
  void load_table(istream &fout, bool tabletype);//true for orf table
  void load_funcats(istream &fin);
  void add_ends(const vector<string> &ss);
  int num_orfs(int c) const {return orf_list[c].size();}
  list<Orf> orfs_chrom(int c) {if(!sorted) sort_orfs();return orf_list[c];}
  list<Orf> orfs_chrom_const(int c) const {return orf_list[c];}
  int num_chroms() const {return num_sequences;}
  void update_map();//could be expensive, only do when necessary, after big updates
  void sort_orfs();//could be expensive, only do once if possible
  Orf find_orf(string yn);
  int free_up(Orf o);
  int free_down(Orf o);
};


class OrfLocale{
  int chrom;
  int pos_l;
  int pos_r;
  Orf this_one;
  Site oloc_site;
  vector<Orf> in_orf;
  vector<Orf> in_other;
  Orf near_left_orf;
  Orf near_right_orf;
  Orf near_left_other;
  Orf near_right_other;
  Orf near_left_orf_c;
  Orf near_right_orf_w;
 public:
  OrfLocale(int s, int n, Orfs &o);
  OrfLocale(Orf f, Orfs &o);//in_orf includes this Orf
  OrfLocale(Site st,int wide, Orfs &o);
  void init(const int s, const int l, const int r, Orfs &o);
  Orf left_orf() const {return near_left_orf;}
  Orf right_orf() const {return near_right_orf;}
  Orf left_other() const {return near_left_other;}
  Orf right_other() const {return near_right_other;}
  Orf left_feat() const{if(near_left_orf.right()>near_left_other.right())
    return near_left_orf;return near_left_other;}
  Orf right_feat() const{if(near_right_orf.left()<near_right_other.left())
    return near_right_orf;return near_right_other;}
  Orf left_up() const {return near_left_orf_c;}
  Orf right_up() const {return near_right_orf_w;}
  vector<Orf> in_orfs() const {return in_orf;}
  vector<Orf> in_others() const {return in_other;}
  int num_inside_coding() const {return in_orf.size();}
  int num_inside_others() const {return in_other.size();}
  int num_inside_features() const {return in_orf.size()+in_other.size();}
  string desc();
  string desc_long();
  bool in_coding() const {if (in_orf.size()!=0||in_other.size()!=0) return true;return false;}
  //considers others as 'coding'
  bool divergent() const {if(in_coding()) return false;if(!left_feat().strand()&&right_feat().strand()) return true;return false;}
  bool convergent() const {if(in_coding()) return false;if(left_feat().strand()&&!right_feat().strand()) return true;return false;}
  bool tandem() const {if(!in_coding()&&!divergent()&&!convergent()) return true;return false;}
  int dist_left_up(){return (pos_l-near_left_orf_c.right());}
  //in these cases, the number returned is the number of shifts required to cause an overlap
  int dist_right_up(){return (near_right_orf_w.left()-pos_r);}
  int dist_left_feat(){return (pos_l-(left_feat()).right());}
  int dist_right_feat(){return ((right_feat()).left()-pos_r);}
};


#endif
